def is_valid_body(body):
    if isinstance(body, dict):
        if ("batch" in body or "content" in body):
            return True
    return False


def is_batch_processing(body):
    return "batch" in body



import json
import os
import shutil
from typing import Any, Dict, List, Optional

import requests

from ..api.api_http import headers
from ..api.api_request import provision_req
from ..api.token_api import TokenAPI
from ..configuration import config
from ..logging import log
from ..util import unwrap
from .build import build
from .coprocessor import Coprocessor
from .decorators import context
from .smartpipe import SmartPipe


def create_coprocessor_docker_file(coprocessor: Coprocessor, save_result: bool) -> None:
    docker_file = f"""FROM python:3.10

ENV SMARTPIPES_PRODUCTION True
ENV COPROCESSOR_ID {coprocessor.id}
ENV SAVE_RESULT_COPROCESSOR {save_result}

WORKDIR /app
COPY . .

RUN pip install --no-cache-dir -r requirements.txt
ENTRYPOINT ["python", "demo.py"]
    """

    if not os.path.exists(f"build/{coprocessor.id}"):
        os.makedirs(f"build/{coprocessor.id}")

    with open(f"build/{coprocessor.id}/Dockerfile", "w") as file:
        file.write(docker_file)

    os.system(
        f"docker buildx build --platform linux/arm64,linux/amd64 -t us-central1-docker.pkg.dev/artifacts-356722/demo/coprocessors/{coprocessor.id}:latest  build/{coprocessor.id}  --push"  # noqa
    )


def create_http_api_entry_point_docker_file() -> None:
    docker_file = """FROM python:3.10

ENV SMARTPIPES_PRODUCTION True
ENV PORT 5000

WORKDIR /app
COPY . .

RUN pip install --no-cache-dir -r requirements.txt
RUN pip install gunicorn


EXPOSE 5000

CMD gunicorn --bind 0.0.0.0:${PORT} --workers 1 --timeout 300 demo:app
    """

    if not os.path.exists("build/http"):
        os.makedirs("build/http")

    with open("build/http/Dockerfile", "w") as file:
        file.write(docker_file)


def create_carrier_workload_file(
    tenant: str, smart_pipe_id: str, coprocessor: Coprocessor, next_coprocessors: List[str]
) -> Dict[str, Any]:
    output: Optional[Dict[str, Any]] = None

    if len(next_coprocessors) > 1:
        output = {
            "broker": {
                "outputs": (
                    {"carrier": {"subject": f"{smart_pipe_id}.{c_id}"}}
                    for c_id in next_coprocessors
                )
            }
        }
    elif len(next_coprocessors) == 1:
        output = {
            "label": "carrier_out",
            "carrier": {"subject": f"{smart_pipe_id}.{next_coprocessors[0]}"},
        }

    workload = {
        "tenant": tenant,
        "id": coprocessor.id,
        "input": {
            "label": "carrier_in",
            "carrier": {
                "subject": f"{smart_pipe_id}.{coprocessor.id}",
                "deliver": "all",
                "queue": coprocessor.id,
            },
        },
        "processor": {
            "docker": {
                "image": f"us-central1-docker.pkg.dev/artifacts-356722/demo/coprocessors/{coprocessor.id}:latest",  # noqa
                "args": [],
            }
        },
        "output": output,
    }

    if not os.path.exists(f"build/{coprocessor.id}"):
        os.makedirs(f"build/{coprocessor.id}")

    with open(f"build/{coprocessor.id}/workload.json", "w") as file:
        json.dump(workload, file, indent=2)
        log.debug(f"Created {coprocessor.id} workload")

    return workload


def copy_project_into_resource(id: str) -> None:
    source_folder = "."
    destination_folder = f"build/{id}"

    if not os.path.exists(f"build/{id}"):
        os.makedirs(f"build/{id}")

    for item in os.listdir(source_folder):
        if os.path.isdir(item) and item == "build":
            continue  # Skip the "build" folder

        elif os.path.isdir(item):
            destination_path = os.path.join(destination_folder, item)
            if os.path.exists(destination_path):
                shutil.rmtree(destination_path)
            shutil.copytree(item, destination_path)
        else:
            destination_path = os.path.join(destination_folder, item)
            shutil.copy2(item, destination_path)


def create_stream(name: str) -> Any:
    log.debug(f"Creating stream: {name}")
    url = f"{config.carrier_endpoint}/stream/{name}"
    req = provision_req(config._token_api)

    payload = {
        "message_ttl": 3600,
        "max_messages": 1000000,
        "max_size": 100000000,
        "replicas": 3,
        "allow_locations": ["region/xn"],
        "deny_locations": ["country/nl"],
        "wait_for_ack": True,
        "ack_timeout": 5,
        "max_delivery_attempts": 5,
        "max_delivery_time": 60,
        "dead_letter_sink": "dead-letter-sink",
    }

    return unwrap(
        req(
            lambda access_token: requests.put(
                url,
                json=payload,
                headers=headers(access_token),
            )
        )
    )


def delete_stream(name: str) -> Any:
    log.debug(f"Deleting stream: {name}")
    url = f"{config.carrier_endpoint}/stream/{name}"
    req = provision_req(config._token_api)

    return unwrap(
        req(
            lambda access_token: requests.delete(
                url,
                headers=headers(access_token),
            )
        )
    )


def create_flow(name: str, workload: Dict[str, Any]) -> Any:
    log.debug(f"Creating flow: {name}")
    url = f"{config.carrier_endpoint}/flow/{name}"
    req = provision_req(config._token_api)

    return unwrap(
        req(
            lambda access_token: requests.put(
                url,
                json=workload,
                headers=headers(access_token),
            )
        )
    )


def delete_flow(name: str) -> Any:
    log.debug(f"Deleting flow: {name}")

    url = f"{config.carrier_endpoint}/flow/{name}"
    req = provision_req(config._token_api)

    return unwrap(
        req(
            lambda access_token: requests.delete(
                url,
                headers=headers(access_token),
            )
        )
    )


def deploy_coprocessor(
    tenant: str, smart_pipe: SmartPipe, coprocessor: Coprocessor, schema: Dict[str, Any]
) -> None:
    delete_flow(coprocessor.id)

    save_result_coprocessor = schema["smartpipes"][smart_pipe.id]["io"].get("returns", None) == coprocessor.id
    copy_project_into_resource(coprocessor.id)
    create_coprocessor_docker_file(coprocessor, save_result_coprocessor)
    next_coprocessors = schema["smartpipes"][smart_pipe.id]["io"].get(coprocessor.id, None)

    if next_coprocessors is None:
        next_coprocessors = []

    workload = create_carrier_workload_file(tenant, smart_pipe.id, coprocessor, next_coprocessors)

    create_flow(coprocessor.id, workload)

    log.info(f"Deploy for coprocessor {coprocessor.id} done")


def deploy(coprocessor_id: Optional[str] = None) -> None:
    schema = build()
    tenant = TokenAPI(config).get_tenant()

    if coprocessor_id is not None and coprocessor_id != "entry_point":
        for sm in context.smart_pipes:
            for c in sm.coprocessors:
                if c.id == coprocessor_id:
                    deploy_coprocessor(tenant, sm, c, schema)
    elif coprocessor_id is not None and coprocessor_id == "entry_point":
        log.info("Deploying entry points...")

        copy_project_into_resource("http")
        create_http_api_entry_point_docker_file()
    else:  # deploy everything
        for sm in context.smart_pipes:
            delete_stream(sm.id)
            create_stream(sm.id)

            for c in sm.coprocessors:
                deploy_coprocessor(tenant, sm, c, schema)

        copy_project_into_resource("http")
        create_http_api_entry_point_docker_file()

    log.info("Deployment complete")


def destroy() -> None:
    build()

    for sm in context.smart_pipes:
        delete_stream(sm.id)

        for c in sm.coprocessors:
            delete_flow(c.id)



from typing import Optional, Any, List, Dict
import psycopg2
from psycopg2.extensions import connection
from psycopg2.extras import Json

class RequestDataSource:
    def __init__(self, database: str, host: str, username: str, password:str, port: int = 5432) -> None:
        self.conn: connection = psycopg2.connect(
            database=database,
            host=host,
            user=username,
            password=password,
            port=port,
        )

        self.conn.set_session(autocommit=True)
    
    def save_request(self, id: str, batch_count: int) -> None:
        cursor = self.conn.cursor()
        cursor.execute(""" INSERT INTO requests
                        (id, status, batch_count)
                        VALUES
                        (%s, %s, %s)""", [id, "processing", batch_count])
        row_count: int = cursor.rowcount
        cursor.close()
        return row_count

    def save_result(self, request_id: str, order: int, output: Any) -> None:
        cursor = self.conn.cursor()
        cursor.execute(""" INSERT INTO results
                        (request_id, original_order, output)
                        VALUES
                        (%s, %s, %s)""", [request_id, order, Json(output)])
        row_count: int = cursor.rowcount
        cursor.close()
        return row_count

    def get_request_status(self, request_id: str) -> Optional[Any]:         
        cursor = self.conn.cursor()
        cursor.execute("""
          SELECT r.id, r.batch_count,
              (SELECT COUNT(*) FROM results rs WHERE rs.request_id = r.id) AS result_count
          FROM requests r
          WHERE r.id = %s;
        """, [request_id])
        row = cursor.fetchone()        
        result = None
        if row:
          result = {
              'request_id': row[0],
              'batch_count': row[1],
              'result_count': row[2]
          }
        cursor.close()
        return result 

    def get_request_results(self, request_id: str) -> List[Dict[str, Any]]:         
        cursor = self.conn.cursor()
        cursor.execute("""
          SELECT output
          FROM results
          WHERE request_id = %s
          ORDER BY original_order ASC;
        """, [request_id])
        rows = cursor.fetchall()        

        results = []
        for row in rows:
            results.append(row[0])        

        cursor.close()
        return results
        


import asyncio
import functools
import json
import os
import sys
import traceback
import uuid

from typing import Any, List, Optional

from flask import Flask, request, jsonify
from flask_cors import CORS
from flask_socketio import SocketIO, emit
import nats

from ..datasource import RequestDataSource
from ...carrier import processor
from ...configuration import config
from ...logging import log
from ...model.errors import SeaplaneError
from ..build import build
from ..decorators import context, smart_pipes_json
from ..deploy import deploy, destroy
from ..smartpipe import SmartPipe
from .request import is_valid_body, is_batch_processing

SMARTPIPES_CORS = list(os.getenv("SMARTPIPES_CORS", "http://localhost:3000").split(" "))
AUTH_TOKEN = os.getenv("SMARTPIPES_AUTH_TOKEN")

app = Flask(__name__)

CORS(app, origins=SMARTPIPES_CORS)

if not config.is_production():

    sio = SocketIO(app, cors_allowed_origins=SMARTPIPES_CORS, async_mode="threading")

    @sio.on("message")  # type: ignore
    def handle_message(data: Any) -> None:
        ...


    @sio.on("connect")  # type: ignore
    def on_connect() -> None:
        emit("message", smart_pipes_json(context.smart_pipes))


def send_something(data: Any) -> None:
    emit("message", data, sid="sp", namespace="", broadcast=True)


def authenticate_token() -> bool:
    token = request.headers.get("Authorization")
    if token and token.startswith("Bearer "):

        if token == f"Bearer {AUTH_TOKEN}":
            return True

    return False


@app.before_request
def before_request() -> Any:
    if request.path == "/healthz":
        return

    if os.getenv("SMARTPIPES_AUTH_TOKEN") is not None:
        if not authenticate_token():
            return {"message": "Unauthorized"}, 401


def dev_https_api_start() -> Flask:
    log.debug(f"CORS enabled: {SMARTPIPES_CORS}")
    context.set_event(lambda data: send_something(data))

    smart_pipes = context.smart_pipes

    for smart_pipe in smart_pipes:

        def endpoint_func(pipe: SmartPipe = smart_pipe) -> Any:
            if request.method == "POST" or request.method == "PUT":
                data = request.get_json()
                result = pipe.func(data)
                return result
            elif request.method == "GET":
                return pipe.func("nothing")  # current limitation, it needs to pass something

        endpoint = functools.partial(endpoint_func, pipe=smart_pipe)
        app.add_url_rule(smart_pipe.path, smart_pipe.id, endpoint, methods=[smart_pipe.method])
        

    def health() -> str:
        emit("message", {"data": "test"}, sid="lol", namespace="", broadcast=True)
        return "Seaplane SmartPipes Demo"

    app.add_url_rule("/", "healthz", health, methods=["GET"])

    if not config.is_production():
        log.info("🚀 Smart Pipes DEVELOPMENT MODE")
        sio.run(app, debug=False, port=1337)
    else:
        log.info("🚀 Smart Pipes PRODUCTION MODE")

    return app


loop = asyncio.get_event_loop()


async def publish_message(stream: str, id: str, message: Any, order: int, coprocessors: List[str]) -> None:
    nc = await nats.connect(
        ["nats://148.163.201.1:2003"],
        user_credentials=os.path.abspath("./carrier.creds"),
    )
    js = nc.jetstream()

    nats_message = { "id": id, "input": message, "order": order }

    for coprocessor in coprocessors:
        log.debug(f"Sending to {stream}.{coprocessor} coprocessor,  message: {nats_message}")
        ack = await js.publish(f"{stream}.{coprocessor}", str(json.dumps(nats_message)).encode())
        log.debug(f"ACK: {ack}")

    await nc.close()


def generate_id():
    random_id = str(uuid.uuid4())
    return random_id

def build_requests_datasource():
    database = os.getenv("SEAPLANE_REQUESTS_DATABASE", None)
    host = os.getenv("SEAPLANE_REQUESTS_HOST", None)
    username = os.getenv("SEAPLANE_REQUESTS_USERNAME", None)
    password = os.getenv("SEAPLANE_REQUESTS_PASSWORD", None)
    port = os.getenv("SEAPLANE_REQUESTS_PORT", None)    
    return RequestDataSource(database=database, host=host, username=username, password=password, port=port)


def prod_https_api_start() -> Flask:
    log.debug(f"CORS enabled: {SMARTPIPES_CORS}")

    schema = build()
    context.set_event(lambda data: send_something(data))        
    requests_datasource = build_requests_datasource()

    smart_pipes = context.smart_pipes

    for smart_pipe in smart_pipes:

        def endpoint_func(pipe: SmartPipe = smart_pipe) -> Any:            
            if request.method == "POST" or request.method == "PUT":
                body = request.get_json()

                if not is_valid_body(body):
                    return jsonify({"error": "Invalid JSON"}), 401

                id = generate_id()
                smart_pipe_first_coprocessors = schema["smartpipes"][pipe.id]["io"]["entry_point"]
                metadata = body["metadata"]

                if is_batch_processing(body):
                    batch = body["batch"]                    

                    for idx, content in enumerate(batch):
                        content["_metadata"] = metadata                        
                        loop.run_until_complete(
                            publish_message(pipe.id, id, content, idx, smart_pipe_first_coprocessors)
                        )
                                        
                    requests_datasource.save_request(id, len(batch))
                else:
                    content = body["content"]
                    content["_metadata"] = metadata                    
                    loop.run_until_complete(
                        publish_message(pipe.id, id, content, 1, smart_pipe_first_coprocessors)
                    )
                    requests_datasource.save_request(id, 1)

                return jsonify({"id": id, "status": "processing"}), 200
            elif request.method == "GET":
                return pipe.func("nothing")  # current limitation, it needs to pass something
        
        def get_result(id: str) -> Any:                        
            
            output = None
            req_status = requests_datasource.get_request_status(id)
            if not req_status: 
                return jsonify({"error": f"Request ID: {id} not found"}), 404

            batch_count = req_status["batch_count"]
            result_count = req_status["result_count"]

            if batch_count != result_count:
                status = f"{result_count}/{batch_count}"
            else:
                status = "completed"        
                output = requests_datasource.get_request_results(id)        
                if len(output) == 1:
                    output = output[0]
            
            return jsonify({"id": id, "status": status, "output": output}), 200

        endpoint = functools.partial(endpoint_func, pipe=smart_pipe)
        app.add_url_rule(smart_pipe.path, smart_pipe.id, endpoint, methods=[smart_pipe.method])
        app.add_url_rule(f"{smart_pipe.path}/<id>", f"{smart_pipe.id}_query", get_result, methods=["GET"])

    def health() -> str:
        return "Seaplane SmartPipes Demo"

    app.add_url_rule("/", "healthz", health, methods=["GET"])

    if not config.is_production():
        log.info("🚀 Smart Pipes DEVELOPMENT MODE")
        sio.run(app, debug=False, port=1337)
    else:
        log.info("🚀 Smart Pipes PRODUCTION MODE")

    return app


def start_coprocessor(coprocessor_id: str, save_result: bool) -> None:
    coprocessor = context.get_coprocessor(coprocessor_id)

    if not coprocessor:
        raise SeaplaneError(
            f"Coprocessor {coprocessor_id} not found, \
                            make sure the Coprocessor ID is correct."
        )

    processor.start()

    requests_datasource = None
    if save_result:
        requests_datasource = build_requests_datasource()

    while True:
        log.info(f" Coprocessor {coprocessor.id}  waiting for getting data...")
        message = json.loads(processor.read())
        log.debug(f" Message recieved: {message}")

        id = message["id"]
        order = message["order"]            
        
        try:                    
            message["output"] = coprocessor.process(message["input"])

            log.debug(f" Coprocessor Result: {message}")

            next_message = {"input": message["output"], "id": id, "order": order}
            processor.write(str(json.dumps(next_message)).encode())

            if requests_datasource:
                requests_datasource.save_result(id, order, message["output"])

        except Exception as e:
            error_str = '\n'.join(traceback.format_exception(type(e), e, e.__traceback__))
            log.error(
                f"Error running Coprocessor:\
                      \n {error_str}"
            )
            next_message = {"input": {"error": error_str}, "id": id, "order": order}
            processor.write(str(json.dumps(next_message)).encode())


def start() -> Optional[Flask]:
    if len(sys.argv) > 1:
        command = sys.argv[1]

        if command == "build":
            build()
            return None
        elif command == "deploy":
            if len(sys.argv) == 3:
                co_id = sys.argv[2]
                deploy(co_id)
            else:
                deploy()
            return None
        elif command == "destroy":
            destroy()
            return None

    coprocessor_id: Optional[str] = os.getenv("COPROCESSOR_ID")

    if not coprocessor_id:
        log.info("Starting API Entry Point...")
        if not config.is_production():
            return dev_https_api_start()
        else:
            return prod_https_api_start()
    else:
        log.info(f"Starting Coprocessor {coprocessor_id} ...")                    
        save_result = os.getenv("SAVE_RESULT_COPROCESSOR").lower() == "true"
        start_coprocessor(coprocessor_id, save_result)
        return None



import json
import os
from typing import Any, Dict

from ..logging import log
from .decorators import context
from .executor import RealCoProcessorExecutor, SchemaExecutor


def persist_schema(schema: Dict[str, Any]) -> None:
    if not os.path.exists("build"):
        os.makedirs("build")

    file_path = os.path.join("build", "schema.json")

    with open(file_path, "w") as file:
        json.dump(schema, file, indent=2)


def build() -> Dict[str, Any]:
    schema: Dict[str, Any] = {"smartpipes": {}}

    context.set_executor(SchemaExecutor())

    for sm in context.smart_pipes:
        result = sm.func("entry_point")
        sm.return_source = result


    for sm in context.smart_pipes:
        smartpipe: Dict[str, Any] = {
            "id": sm.id,
            "entry_point": {"type": "API", "path": sm.path, "method": sm.method},
            "coprocessors": [],
            "io": {},
        }

        for c in sm.coprocessors:
            coprocessor = {"id": c.id, "name": c.name, "type": c.type, "model": c.model}

            for source in c.sources:
                if not smartpipe["io"].get(source, None):
                    smartpipe["io"][source] = [c.id]
                else:
                    smartpipe["io"][source].append(c.id)
            
            smartpipe["coprocessors"].append(coprocessor)

        smartpipe["io"]["returns"] = sm.return_source
        schema["smartpipes"][sm.id] = smartpipe

    persist_schema(schema)
    log.debug("Created Smart Pipes configuration")

    context.set_executor(RealCoProcessorExecutor(context.event_handler))

    return schema
